/** ---------------------------------------------------------------
@mainpage W5UXH Keyer Implemented on mbed NXP LPC1768
@version x.y
@author Chuck Broadwell, W5UXH  Las Cruces, NM
@date Begun April, 2012
@verbatim

A keyboard/keyer program for CW running on the mbed NXP LPC1768

A USB keyboard is used for input, with an option for serial input over UART0.
A 4x20 LCD module using the HD44870 controller is used for display.

The features are based on an earlier (circa 1999) PIC16F876 design and a more recent (circa 2010) 
ATmega328P design, both by W5UXH . 

A "ten minute" timer for ID is provided that resets to zero whenever the user's call is sent from 
either the PS-2 keys or from the paddle.  These features go back to older hardware versions and 
early PIC versions.

    Copyright (C) 2012


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


--------------------------------------------------------------------------------------
@endverbatim
**/

#include "mbed.h"
#include "TextLCD.h"
#include "w5uxhMbedKeyer.h"
#include "queues.h"
#include <ctype.h>

// Pointers to the FIFO objects instantiated as statics in keyerInit()
Fifo *cwBuf;
Fifo *lcdBuf;
Fifo *consoleBuf;
Fifo *cmdBuf;

int cwBufSize;

PwmOut sidetone(p21);

extern unsigned int kbdDotLength_usec;
extern unsigned int kbdDashLength_usec;
extern unsigned int kbdElSpace_usec;
extern unsigned int kbdCharSpace_usec;
extern unsigned int kbdWordSpace_usec;

extern void (*ctrlFuncs[])(char);

char idString[10];
char idDetected = FALSE;
int idDetectCounter = 0;
int idStringLength;
int idDetectCounterPdl = 0;
extern unsigned char newPaddleChar;
extern unsigned char pdlCharSpaceDetected;
extern unsigned char pdlWordSpaceDetected;

unsigned char kbdWPM=60;
unsigned char pdlWPM=30;
unsigned char weight = 52;
unsigned char asciiToMorseLookup (unsigned char ascii);

TextLCD lcd(p15, p16, p17, p18, p19, p20, TextLCD::LCD20x4); // rs, e, d4-d7

extern Serial pc;

DigitalOut led1(LED1);

extern Timeout cwTimer;
extern void atCwTimer();
extern void startCwCharacter(unsigned char morseByte);
extern unsigned char kbdCwState;

Ticker idTimer;
int timeTicIntrCount = 0;
int minutes=8;
int seconds=55;
int newHalfSecondFlag = FALSE;

#define ESCAPE_START    1
#define LOOK_FOR_THIRD  2
#define LOOK_FOR_FOURTH 3

int escState = IDLE;
bool speedChangeFlag = FALSE;

bool newCharFlag;
char newChar;
char ctrlCode;

//******************************************************************************
// This is from http://mbed.org/forum/mbed/topic/3179/?page=1#comment-16243

// These external symbols are maintained by the linker to indicate the
// location of various regions in the device's memory.  They will be used by
// DisplayRAMBanks() to dump the size of each RAM bank to stdout.
extern unsigned int Image$$RW_IRAM1$$Base;
extern unsigned int Image$$RW_IRAM1$$ZI$$Limit;
extern unsigned int Image$$RW_IRAM2$$Base;
extern unsigned int Image$$RW_IRAM2$$ZI$$Limit;
extern unsigned int Image$$RW_IRAM3$$Base;
extern unsigned int Image$$RW_IRAM3$$ZI$$Limit;


// Displays the size of static allocations for each RAM bank as indicated by
// ARM linker to stdout.
#if 0
static void DisplayRAMBanks(void)
{
    pc.printf("Static RAM bank allocations\r\n");
    pc.printf("  Main RAM = %u\r\n", (unsigned int)&Image$$RW_IRAM1$$ZI$$Limit - 
                                  (unsigned int)&Image$$RW_IRAM1$$Base);
    pc.printf("  RAM0     = %u\r\n", (unsigned int)&Image$$RW_IRAM2$$ZI$$Limit -
                                  (unsigned int)&Image$$RW_IRAM2$$Base);
    pc.printf("  RAM1     = %u\r\n", (unsigned int)&Image$$RW_IRAM3$$ZI$$Limit -
                                  (unsigned int)&Image$$RW_IRAM3$$Base);
}
#endif
//******************************************************************************


/** -------------------------------------------------------------------------
@verbatim
Update the LCD display of the ID timer every half second.

@endverbatim
**/
void  handleTimerDisplay(void)
{
    if (idDetected)
    {
      idDetected = FALSE;
      __disable_irq();
        minutes = 0;  seconds = 0;
      __enable_irq();
      displayTimerSerial(minutes, seconds, timeTicIntrCount);
      displayTimerLCD(minutes, seconds, timeTicIntrCount);
    }

  if (newHalfSecondFlag)
    {
      newHalfSecondFlag = FALSE;

      if (seconds == 60)
    {
      minutes++;
      seconds = 0;
      if (minutes == 100)
        {
          minutes = 99;
        }
      
    }
      displayTimerSerial(minutes, seconds, timeTicIntrCount);
      displayTimerLCD(minutes, seconds, timeTicIntrCount);
    }
}

/** -------------------------------------------------------------------------
@verbatim
ID timer.  Interrupts every 500 msecs.  This allows us to control the time display
format on half second intervals for the 8 through 12 minute warnings that it is time to ID.
newSecondsFlag is checked in the main idle loop and the display is updated as needed.
@endverbatim
**/
void atEachTicker1(void)
{
  timeTicIntrCount++;
  newHalfSecondFlag = TRUE;

  if (timeTicIntrCount == 2)
    {
      seconds++;
      timeTicIntrCount = 0;
    }
}


/** -------------------------------------------------------------------------
@verbatim
Interrupt handler for the pc UART input data.  Needs a lot of work to move things
outside of the interrupt handler.

@endverbatim
**/
void rxDataRcvd(void)
{
  char inChar;
  bool bufWriteStatus=FALSE;
    
  inChar = pc.getc();

#if 0
  // only use this to identify key code sequences
  pc.printf("%02X ", inChar);
  if (inChar == 0x1B)
    {
      pc.printf("(ESC) ");
    }
  else
    {
      pc.printf("(%c) ", inChar);
    }
#endif

  if (inChar == 0x0D) inChar =0x3D;

#if 0
  THIS HAS GOTTEN PRETTY NASTY AND WILL NEED SOME REAL WORK TO CLEAN IT UP!!!!
#endif   

    if (escState == LOOK_FOR_FOURTH)
    {
      // We invoke the function as if the code came from the USB keyboard handler.

      (*ctrlFuncs[ctrlCode-0x80])(ctrlCode);
      escState = IDLE;
    }


    if (escState == LOOK_FOR_THIRD)
    {
      switch(inChar)
    {
    case 'A': ctrlCode = 0xB9; speedChangeFlag = TRUE; break;
    case 'B': ctrlCode = 0xBB; speedChangeFlag = TRUE; break;
    case 'C': ctrlCode = 0xBC; speedChangeFlag = TRUE; break;
    case 'D': ctrlCode = 0xBA; speedChangeFlag = TRUE; break;
      //For the next four key codes, we need to LOOK_FOR_FOURTH which should
      //be the tilde character.  We throw it away and then use ctrlCode
    case '2': ctrlCode = 0xB2; escState = LOOK_FOR_FOURTH; break;  // EQUAL (INSERT)
    case '3': ctrlCode = 0xB5; escState = LOOK_FOR_FOURTH; break;  // PDL (Delete)
    case '4': ctrlCode = 0xB6; escState = LOOK_FOR_FOURTH; break;  // KBD (End)
    case '6': ctrlCode = 0xB7; escState = LOOK_FOR_FOURTH; break;  // WT  (PgDn)
    }
      // If escState still is LOOK_FOR_THIRD, then we are done, otherwise we have changed
      // escState to LOOK_FOR_FOURTH above.
      if (escState == LOOK_FOR_THIRD)
    {
      // Either we have found one of the characters or it was some key we do not handle.
      if (speedChangeFlag == TRUE)
        {
          (*ctrlFuncs[ctrlCode-0x80])(ctrlCode);
        }
      escState = IDLE;
    }
    }

  if (escState == ESCAPE_START)
    {
      if (inChar == '[')
    {
      escState = LOOK_FOR_THIRD;
    }
      else
    {
      escState = IDLE;
    }
    }

  if (inChar == 0x1B)
    {
      escState = ESCAPE_START;
    }

  // Only backspace if there is still data in the buffer.
  if ((inChar == 0x08) && (cwBuf->avail()))
    {
      cwBuf->backspace();
    }
  else
    {
      if (inChar == 0x18)
    { 
      cwBuf->reset();  //ctrl-X
      dsplyBufStatusSerial(NO_FORCE);
      dsplyBufStatusLCD();
    }         
      else
    {
      if ((!speedChangeFlag) && (escState == IDLE) && (inChar != 0x08) && (inChar != '~'))
        {
          bufWriteStatus = cwBuf->write(inChar);
          if (inChar == '=') bufWriteStatus = cwBuf->write(' ');
        }
    }
    }

  // eventually will need to put the characters into a buffer to be displayed
  // outside the interrupt!!!!!!!!!!
  if ((!speedChangeFlag) && (escState != LOOK_FOR_THIRD) && (escState != ESCAPE_START)   )
    {
      //newCharFlag = TRUE;
      //newChar = inChar;

      // Put char in buffer to be displayed in the idle loop
      // if it is not the case that the char is backspace AND the cwBuf is empty

      if (inChar != 0x08)
    {
      if ((bufWriteStatus) && (inChar != '~')) consoleBuf->write(inChar);
    }
      else
    {
      if (cwBuf->avail()) consoleBuf->write(inChar);
    }
    }
  // Clear flag in case it was set.
  speedChangeFlag = FALSE;
}


/** -------------------------------------------------------------------------
@verbatim
Display new character from paddle

@endverbatim
**/
void displayPdlChar(char outChar)
{
  char c[3];

  c[0] = '?';
  c[1] = '?';
  c[2] = 0;

  if ((outChar >= 0x20) && (outChar < 0x80))
    {
      displayLcdTextRow(2,outChar);
      pc.printf("%c", outChar);
    }
  else
    {
      switch (outChar)
    {
    case AS: c[0]='A'; c[1]='S';  break;
    case VE: c[0]='V'; c[1]='E';  break;
    case AR: c[0]='A'; c[1]='R';  break;
    case SK: c[0]='S'; c[1]='K';  break;
    case AA: c[0]='A'; c[1]='A';  break;
    case BT: c[0]='B'; c[1]='T';  break;
    }      
      displayLcdStringRow(2, c);
      pc.printf("%s", c);
    }
}

/** -------------------------------------------------------------------------
@verbatim
Display characters received from the console.
Handle one char per pass through idle loop.

@endverbatim
**/
void handleCharFromConsole(void)
{
  char consoleChar;

  if (consoleBuf->avail())
    {
      consoleChar = consoleBuf->read();
      if (consoleChar == '=')
    {
      pc.printf(" BT\r");
      displayLcdTextRow(2,' ');
      displayLcdTextRow(2,'B');
      displayLcdTextRow(2,'T');
      displayLcdTextRow(2,' ');
    }
      else
    {
          pc.printf("%c", consoleChar);
        //  checkId(consoleChar);  // moved to output of cwBuf
      if (consoleChar == 0x08) pc.printf(" \b");
          displayLcdTextRow(2,consoleChar);      
    }
    }
}


extern unsigned char arrowsControl;
extern unsigned char allowAutoArrowKeyControl;
extern unsigned char toggleSameSpeedModeState;
/** -------------------------------------------------------------------------
@verbatim
Check to see if we need to change the arrow key focus based on input (paddle or kbd)

@endverbatim
**/
void checkAutoFocus(int inputDevice)
{
#define IS_KBD 0xB6
#define IS_PDL 0xB5

  if ((allowAutoArrowKeyControl) && (toggleSameSpeedModeState == IDLE))
    {
      if (inputDevice == KBD)
	{
	  if (arrowsControl != KBD)
	    {
	      arrowsControl = KBD;
	      setFocus(IS_KBD);
	    }
	}
      else
	{
	  // inputDevice is PDL
	  if (arrowsControl != PDL)
	    {
	      arrowsControl = PDL;
	      setFocus(IS_PDL);
	    }

	}
    }
}

/** -------------------------------------------------------------------------
@verbatim
Check for Id from keyboard

@endverbatim
**/
void checkId(char newAsciiChar)
{
  if (toupper(idString[idDetectCounter]) == toupper(newAsciiChar))
    {
      idDetectCounter++;
      if (idDetectCounter == idStringLength)
      {
        idDetected = TRUE;
        idDetectCounter = 0;
      }
    }
  else
    {
      idDetectCounter = 0;
    }
}

/** -------------------------------------------------------------------------
@verbatim
Check for Id from paddle

@endverbatim
**/
void checkPaddleId(char newAsciiChar)
{
  if (toupper(idString[idDetectCounterPdl]) == toupper(newAsciiChar))
    {
      idDetectCounterPdl++;
      if (idDetectCounterPdl == idStringLength)
      {
        idDetected = TRUE;
        idDetectCounter = 0;
      }
    }
  else
    {
      idDetectCounterPdl = 0;
    }
}

void getCompileTime(void)
{
  char tempString[20];
  char month[4];
  char day[3];
  char year[3];

  pc.printf("%s %s\r", __DATE__, __TIME__);

  strcpy(tempString, __DATE__);
  month[0] = tempString[0];
  month[1] = tempString[1];
  month[2] = tempString[2];
  month[3] = 0;

  day[0] = tempString[4];
  day[1] = tempString[5];
  day[2] = 0;

  year[0] = tempString[9];
  year[1] = tempString[10];
  year[2] = 0;

  lcd.locate(0,3);
  lcd.printf("%s%s%s-%s", day, month, year, __TIME__ );
}


int nrOfRows = 40;

/** -------------------------------------------------------------------------
@verbatim
Console initialization

@endverbatim
**/
void initConsole(void)
{
  clearSerialScreen();
  enableSerialLineWrap();
  setConsoleScrollRows(FIRST_DISPLAY_ROW, FIRST_DISPLAY_ROW + nrOfRows);
  displaySpeedsSerial();

  setCursor(FIRST_DISPLAY_ROW,1);
  displayTimerSerial(minutes, seconds, timeTicIntrCount);
  dsplyBufStatusSerial(FORCE);
  
    switch(nrOfRows)
    {
    case 40: pc.printf("40 lines of text\r"); break;
    case 4:  pc.printf("4 lines of text\r"); break;
    case 2:  pc.printf("2 lines of text\r"); break;
    }
}


/** -------------------------------------------------------------------------
@verbatim
keyerInit will be called from the main function in the USB section.

@endverbatim
**/
void keyerInit(void) 
{
  int baudrate = PC_BAUD;
  char tempString[20];

  // I don't know if I will need to play with interrupt priorities or not.  I believe the
  // default priorities for everything are all highest (0)
  NVIC_SetPriority(TIMER3_IRQn, 2);
  NVIC_SetPriority(UART1_IRQn, 1);
  // USB_IRQn is set in USBHost.cpp

  // I will keep this function here for awhile at least, to see what happens to SRAM
  // usage as things are added.
  
  //  DisplayRAMBanks();
  
  // 620 Hz sidetone
  sidetone.period_us(1613);

  idTimer.attach_us(&atEachTicker1, 500000);

  pc.attach(&rxDataRcvd);
  

  calcTimeDurationsKbd(kbdWPM, weight);
  calcTimeDurationsPdl(pdlWPM, weight);

  static  Fifo cwBufStatic(2048, "cwBuf");
  static  Fifo lcdBufStatic(64, "lcdBuf");
  static  Fifo consoleBufStatic(64, "consoleBuf");
  static  Fifo cmdBufStatic(64, "cmdBuf");

  cwBuf      = &cwBufStatic;
  lcdBuf     = &lcdBufStatic;
  consoleBuf = &consoleBufStatic;
  cmdBuf     = &cmdBufStatic;



  // Store size as a global for use with displays of buf status.
  cwBufSize = cwBuf->get_size();

  initConsole();
  fileInputSettings();
  idStringLength = strlen(idString);
  


  displaySpeedsLCD();
  dsplyBufStatusLCD();
  displayConsoleID();

  getCompileTime();
 
  //  lcd.locate(0,3);
  //lcd.printf("                    ");
  //lcd.locate(0,3);
  //lcd.printf(idString);

  displayLcdStringRow(3, idString);

  int pad;
  int i;
  pad = 19-strlen(idString);
  for(i=0; i<pad; i++)
    {
      displayLcdTextRow(3, ' ');
    }
  
  #ifdef USE_SERIAL_OVER_UART1
  sprintf(tempString, "Using UART1: %5dB", baudrate);
  displayLcdStringRow(2, tempString);
  #endif

  #ifdef USE_SERIAL_OVER_UART2
  sprintf(tempString, "Using UART2: %5dB", baudrate);
  displayLcdStringRow(2, tempString);
  #endif  

  #ifdef USE_SERIAL_OVER_UART3
  sprintf(tempString, "Using UART3: %5dB", baudrate);
  displayLcdStringRow(2, tempString);
  #endif
  
  #ifdef USE_SERIAL_OVER_USB
  sprintf(tempString, "UART_USB:    %5dB", baudrate);
  displayLcdStringRow(2, tempString);
  #endif


}




/** -------------------------------------------------------------------------
@verbatim
keyerIdleLoop will be called from the USBLoop function in the USB section.

@endverbatim
**/
void keyerIdleLoop(void)
{
  unsigned char newPdlAsciiChar;

  if (kbdCwState == KBD_IDLE)
    {
      checkCwBuffer();
    }

  handleCharFromConsole();
  handleTimerDisplay();

  if (cwBuf->getStatusFlag())
    {
      cwBuf->clearStatusFlag();
      dsplyBufStatusLCD();
      dsplyBufStatusSerial(NO_FORCE);
    }
  checkPaddles();

  if(pdlCharSpaceDetected == TRUE)
    {
      pdlCharSpaceDetected = IDLE;
      newPdlAsciiChar = morseToAsciiLookup(newPaddleChar);
      checkPaddleId(newPdlAsciiChar);
      displayPdlChar(newPdlAsciiChar);

      //      displayLcdTextRow(2, newPdlAsciiChar);
      //      pc.printf("%c", newPdlAsciiChar);
    }

  if(pdlWordSpaceDetected == TRUE)
    {
      pdlWordSpaceDetected = IDLE;
      displayLcdTextRow(2, ' ');
      pc.printf("%c", ' ');
    }
}
